#ifdef _WIN32
  #ifndef STRICT
    #define STRICT
  #endif
  #include <windows.h>
  #include <winbase.h>
  #include <winsock.h>
  #include <stdio.h>
  #include <io.h>
  #include <sys\timeb.h>
  #include <time.h>
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <direct.h>
  #include <fcntl.h>
#else
  #include <sys/types.h>
  #include <netdb.h>
  #include <fcntl.h>
  #include <stdarg.h>
  #include <sys/timeb.h>
  #include <sys/time.h>
  #include <sys/stat.h>
  #include <sys/socket.h>
  #include <sys/wait.h>
  #include <netinet/in.h>
  #include <netinet/tcp.h>
  #include <unistd.h>
  #include <stdio.h>
  #include <dirent.h>
  #include <errno.h>
  #include <sys/file.h>
  #include <sys/termios.h>
  #include <arpa/inet.h>
  #include <string.h>
  #include <stdlib.h>
  #define BYTE uint8_t
#endif
#include <stdio.h>
#include <string>
#include <vector>
#include <chrono>
#include <thread>

class Serial
{
  // ThMutex   aMtx;
public:
  std::string          sName;
  std::vector<uint8_t> vReadBuf;
  bool   bOpen;
#ifdef _WIN32
  HANDLE CommId;
#else
  FILE * _fp_serial;
  int    _fd_serial;
  struct termios _tser;
  struct termios _tsero;
  fd_set readseta;
#endif
  volatile int bBlockWriteSerial;

       Serial();
      ~Serial();
  
  int64_t Open(const char *ser_dev, int BitRate);
  void    Close();
  int     Read(BYTE *buf, int maxlen, int wait);
  int     Write(BYTE *buf, int maxlen, bool bWait=false);
  int     Write(const char *pc);
  char*   Read(int maxlen, int wait, int &readlen);
  void    Flush();
};


int bDebug=0;

Serial::Serial()
{
  // aMtx.Create();

  bBlockWriteSerial=0;

  bOpen=false;
  #ifdef _WIN32
    CommId=INVALID_HANDLE_VALUE;
  #else
    _fp_serial=0;
    _fd_serial=0;
  #endif
}

Serial::~Serial()
{
  if (bOpen)
    Close();
}

int64_t Serial::Open(const char *ser_dev, int BitRate)
{
#ifdef _WIN32
  // Schnittstelle ffnen
  char devname[64];
  snprintf(devname, sizeof(devname), "\\\\.\\%s", ser_dev);
  CommId=CreateFile(devname, 
                    GENERIC_READ | GENERIC_WRITE,
                    0,
                    NULL,
                    OPEN_EXISTING,
                    FILE_ATTRIBUTE_NORMAL,
                    (HANDLE) NULL); 
     
  if (CommId == INVALID_HANDLE_VALUE)
    return -1;
 
  // Timeouts einstellen
  COMMTIMEOUTS  aTout;
  BOOL b=GetCommTimeouts(CommId, &aTout);

  aTout.ReadIntervalTimeout=1000;
  aTout.ReadTotalTimeoutMultiplier=0;
  aTout.ReadTotalTimeoutConstant=10000;
  aTout.WriteTotalTimeoutMultiplier=100;
  aTout.WriteTotalTimeoutConstant=2000;

  // Sende und Empfangsbuffer setzen
  SetupComm( CommId, 5120, 5120);

  // Alle eventuellen Operationen beenden, in-out-Buffer leeren
  PurgeComm( CommId, PURGE_TXABORT | PURGE_RXABORT |
                     PURGE_TXCLEAR | PURGE_RXCLEAR);

  DCB     aDCB;

  aDCB.DCBlength = sizeof(DCB);
  GetCommState(CommId, &aDCB ) ;

  aDCB.BaudRate=BitRate;
  aDCB.ByteSize=8;
  aDCB.Parity=0;
  aDCB.StopBits=0;
  
  b=SetCommState(CommId, &aDCB); 
  b=SetCommTimeouts(CommId, &aTout);

  // Alle eventuellen Operationen beenden, in-out-Buffer leeren
  PurgeComm( CommId, PURGE_TXABORT | PURGE_RXABORT |
                     PURGE_TXCLEAR | PURGE_RXCLEAR);

  if (CommId != INVALID_HANDLE_VALUE)
    bOpen=true;
  return (int64_t) CommId;
#else
  int ser;

// if (hec_lock(ser_dev)) return(-1);
  ser=open(ser_dev,O_RDWR|O_NDELAY|O_EXCL);
  if (ser==-1) 
    {
    //if (_lckfile[0]) unlink(_lckfile);
    return -1;
    }

   char caSpeed[32];
   snprintf(caSpeed, sizeof(caSpeed), "B%i", BitRate);
   tcgetattr(ser,&_tsero);
  _tser=_tsero;
  _tser.c_iflag &= ~(IGNBRK | IGNCR | INLCR | ICRNL | IUCLC | IXANY | IXON | IXOFF | INPCK | ISTRIP);
  _tser.c_iflag |= (BRKINT | IGNPAR | IGNBRK);
  _tser.c_oflag &= ~OPOST;
  _tser.c_lflag &= ~(ICANON | ISIG | ECHO | ECHONL | ECHOE | ECHOK);
  _tser.c_cflag &= ~(HUPCL | CRTSCTS | CSIZE ) ;
  _tser.c_cflag |= CREAD  | CS8 | CSTOPB | CLOCAL; // | CRTSCTS
  _tser.c_cc[VMIN] = 1;
  _tser.c_cc[VTIME] = 0;
  cfsetospeed(&_tser, B115200);
  cfsetispeed(&_tser, B115200);
  tcsetattr(ser, TCSANOW, &_tser);
  _fd_serial = ser;
  tcflush(_fd_serial,TCIOFLUSH);
  _fp_serial = fdopen(ser, "w+");
  bOpen=true;
  return(ser);
#endif
}

/********************************************************************
 * close serial port and delete lockfile
 * Parameters: fd for pbx
 ********************************************************************/
void Serial::Close()
{
#ifdef _WIN32
  if (CommId != INVALID_HANDLE_VALUE)
    CloseHandle(CommId);
  CommId=INVALID_HANDLE_VALUE;
  bOpen=false;
#else
  int serori = 0;
  fcntl(_fd_serial,F_SETFL,serori);
  _tser = _tsero;
  tcflush(_fd_serial,TCIOFLUSH);
  tcsetattr(_fd_serial,TCSANOW,&_tser);

  close(_fd_serial);
  //if (_lckfile[0]) unlink(_lckfile);
  bOpen=false;
 return;
#endif
}

void Serial::Flush()
{
  if (!bOpen)
    return;
#ifdef _WIN32
  PurgeComm( CommId, PURGE_TXABORT | PURGE_RXABORT |
                     PURGE_TXCLEAR | PURGE_RXCLEAR);
#else
  tcflush(_fd_serial,TCIOFLUSH);
#endif
  vReadBuf.clear();
}

int Serial::Read(BYTE *buf, int maxlen, int wait)
{
#ifdef _WIN32
  DWORD   BytesRead=0;
  DWORD   Errors;
  COMSTAT aComStat;

  BOOL b;
  COMMTIMEOUTS  aTout;
  b=GetCommTimeouts(CommId, &aTout);
  aTout.ReadIntervalTimeout=0;
  aTout.ReadTotalTimeoutMultiplier=0;
  aTout.ReadTotalTimeoutConstant=wait;
  b=SetCommTimeouts(CommId, &aTout);
  b=ReadFile(CommId, buf, maxlen, &BytesRead, 0);
  return (int) BytesRead;
#else
  struct timeval timeout;
  int ret=0;
  int nsel=0;

  errno = 0;

  if (wait >= 1000) 
    {
    timeout.tv_sec = wait / 1000; 
    timeout.tv_usec= 0;
    } 
  else 
    {
    timeout.tv_sec = 0; 
    timeout.tv_usec= wait*1000;
    }

  FD_ZERO(&readseta);
  FD_SET(_fd_serial, &readseta);
  if ((nsel=select(_fd_serial + 1, &readseta, NULL, NULL, (wait) ? &timeout : NULL))) 
    {
    if (FD_ISSET(_fd_serial, &readseta)) 
      {
      ret=read(_fd_serial, buf, maxlen);
      if (!ret) errno = ENOTCONN;
        {
        return (ret);
        }
      }
    }

  return(ret);
#endif
}

int Serial::Write(BYTE *buf, int len, bool bWait)
{
  // Im ListenOnly Modus kein Schreiben in die Schnittstelle mehr erlauben
//  ThMutexLocker aMl(&aMtx,1);
  if (bBlockWriteSerial)
    return len;
  
#ifdef _WIN32
  DWORD Wrote=0;
  if (!WriteFile(CommId, buf, len, &Wrote, NULL))
    return -1;
  return (int) Wrote;
#else
  static fd_set writeseta;
  struct timeval timeout;
  int ret=0;
  errno = 0;
  timeout.tv_sec = 2;
  timeout.tv_usec = 0;
  
  FD_ZERO(&writeseta);
  FD_SET(_fd_serial, &writeseta);

  select(_fd_serial + 1, NULL, &writeseta, NULL, &timeout);
  if (FD_ISSET(_fd_serial, &writeseta)) 
    {
    ret=write(_fd_serial, buf, len);
    if (bWait)
      tcdrain(_fd_serial);
    return(ret);
    }
  return(-1);
#endif
}

int Serial::Write(const char *pc)
{
  return Write((BYTE*)pc, strlen(pc));
}

char * Serial::Read(int maxlen, int waitms, int &readlen)
{
  readlen=0;
  if (maxlen<=0 || maxlen>16384)
    return 0;  

  vReadBuf.resize((size_t)maxlen);
  int ReadLen=Read((BYTE*)vReadBuf.data(), vReadBuf.size(), waitms);
  if (ReadLen<vReadBuf.size())
    vReadBuf.resize(ReadLen);
  readlen=ReadLen;  
  return (char*)vReadBuf.data();  
}

void delay_ms(int x)
{
  std::this_thread::sleep_for(std::chrono::milliseconds(x));
}

int main(int argc, const char *argv[])
{
  if (argc<2)
    {
    perror("using usbcdctest comport\n");
    return -1;
    }
    
  Serial aSer;  
  
  if (aSer.Open(argv[1], 115200)<=0)
    {
    perror("com open error\n");
    return -1;
    }
  
  BYTE wrbuf[1024];
  BYTE rdbuf[1024];
  memset(wrbuf, 'x', sizeof(wrbuf));
  
  int len=1;
  while (1)
    {
    if (len%64==0) 
      { 
      printf("please press enter:\n");
      getchar();
      }
  
    for (int n=0; n<len; n++)
      wrbuf[n]='a'+n%26;
    
    int wrlen=0;
    while (wrlen<len)
      {
      int wl=len-wrlen;
      if (wl>32)
        wl=32;
      int wr=aSer.Write(wrbuf+wrlen, wl);
      wrlen+=wr;
      }
    int wr=wrlen;
    wrbuf[wr]=0;
    printf("send %i bytes %s\n", wrlen, "");//wrbuf); 
    
    int rlen=0;
    int n=0;
    for (n=0; n<200 && rlen<wr; n++)
      {
      int nread=aSer.Read(rdbuf+rlen, len-rlen, 10);
      if (nread>0)
        printf("  -read %i bytes\n", nread);
      if (nread>=0)
        rlen+=nread;
      }        
    if (n==200)
      {
      printf("timeout, read %i bytes\n", rlen);
      return -1;
      }      
      // delay_ms(300);
    if (memcmp(wrbuf,rdbuf,wr)!=0)
      printf("compare error\n");
    if (wr>=0)
      {
      rdbuf[wr]=0;
      printf("read %i bytes %s\n", wr, rdbuf);   
      }
    len++;
    if (len>=512)
      len=1;    
    }
    
  return 0;
}  
  
